
/*
 * mod_mcs.c: Codetable translation
 * 
 * Eugene G. Crosser <crosser@average.org>
 * 
 */

#include "httpd.h"
#include "http_config.h"

#include "libtrans.h"

#include <stdarg.h>

#define MAX_TABNAME 3

typedef struct {
    array_header *langs;
    table *charsets;
    table *agents;
    char *force;
} mcs_dir_conf;

module mcs_module;

/**************** Configuration handling *********************/

static void *create_mcs_config (pool *p, char *dummy)
{
    mcs_dir_conf *a =
      (mcs_dir_conf *)pcalloc (p, sizeof(mcs_dir_conf));

    a->langs = make_array (p, 4, sizeof(char*));
    a->charsets = make_table (p, 4);
    a->agents = make_table (p, 4);
    a->force=NULL;
    return a;
}

static void *merge_mcs_config (pool *p, void *basev, void *addv)
{
    mcs_dir_conf *a =
	(mcs_dir_conf *)pcalloc (p, sizeof(mcs_dir_conf));
    mcs_dir_conf *base = (mcs_dir_conf *)basev,
	*add = (mcs_dir_conf *)addv;

    a->langs = append_arrays (p, add->langs, base->langs);
    a->charsets = overlay_tables (p, add->charsets, base->charsets);
    a->agents = overlay_tables (p, add->agents, base->agents);
    if (add->force) a->force = add->force;
    return a;
}

static char *add_tranlang(cmd_parms *cmd, mcs_dir_conf *s, char *l)
{
    char **a=(char**)push_array(s->langs);
    *a=pstrdup(cmd->pool,l);
    return NULL;
}

static char *add_trancs(cmd_parms *cmd, mcs_dir_conf *s, char *f, char *r)
{
    table_set (s->charsets,r,f);
    return NULL;
}

static char *add_tranua(cmd_parms *cmd, mcs_dir_conf *s, char *f, char *r)
{
    table_set (s->agents,r,f);
    return NULL;
}

static char *add_force(cmd_parms *cmd, mcs_dir_conf *s, char *l)
{
    s->force=pstrdup(cmd->pool,l);
    return NULL;
}

static command_rec mcs_cmds[] = {
{ "TranslateLanguage", add_tranlang, NULL, OR_FILEINFO, ITERATE, 
    "languages that should be translated"},
{ "TranslateCharset", add_trancs, NULL, OR_FILEINFO, ITERATE2, 
    "translation name and charset names"},
{ "TranslateUserAgent", add_tranua, NULL, OR_FILEINFO, ITERATE2, 
    "translation name and user agent substrings"},
/* ForceTranslation would not work...  Need some changes in libtrans... */
{ "ForceTranslation", add_force, NULL, OR_FILEINFO, TAKE1, 
    "force pre-translation (usually directory-dependant)"},
{ NULL }
};

/********************** Translation ***********************/

extern void *(*btranslate)(void *buf, int nbytes);

/* We cannot translate in place because sometimes constant data (placed
   in the (read-only) code segment needs to be written out... */

static int bufsize=0;
static char *buffer=NULL;

static void *translate_buffer(void *buf, int nbytes) {
	if (buf == NULL) return buf;
	if (nbytes > bufsize) {
		free(buffer);
		bufsize=nbytes;
		buffer=(char*)malloc(bufsize);
	}
#ifdef MCS_DEBUG
	printf("will translate %d bytes at %08lx\n",nbytes,buf);
#endif
	while (--nbytes >= 0) buffer[nbytes]=TR_OUT(((char*)buf)[nbytes]);
#ifdef MCS_DEBUG
	printf("translated\n");
#endif
	return buffer;
}

/*********************** Handlers *************************/

static void initialize_mcs (server_rec *s, pool *p) {

    return;
}

/****************** Translation setup *********************/

int setup_translation (request_rec *r)
{
    char trname[4],*p;
    int i,do_translate=0;
    table *e = r->subprocess_env;
    mcs_dir_conf *conf =
	(mcs_dir_conf *)get_module_config(r->per_dir_config, &mcs_module);
    array_header *langs=conf->langs;
    table *charsets=conf->charsets;
    table *agents=conf->agents;
    char *force=conf->force;
    char *acs=table_get(r->headers_in,"Accept-charset");
    table_entry *ll;
    char *ccs=NULL;

#ifdef MCS_DEBUG
    printf ("check: %s \"%s\",\"%s\" and reset (%s)\n",
		r->server->server_hostname,
		r->content_language?r->content_language:"(NULL)",
		r->content_type?r->content_type:"(NULL)",
		r->filename?r->filename:"(NULL)");
#endif
    /* reset translation */
    btranslate=NULL;
    settrtab_byname("nul");
    trname[0]='\0';

    if (r->content_language) {
	char **ll=(char**)langs->elts;
	for (i=0;i<langs->nelts;i++) {
	    if (strcasecmp(ll[i],r->content_language) == 0) {
#ifdef MCS_DEBUG
		printf("allowed language: %s found\n",r->content_language);
#endif
		do_translate=1;
		break;
	    }
	}
#ifdef LOCAL_ROL_KLUDGE
    /* by local conventions rhtmls and rcgis need translation */
    } else {
	if ((strcmp(r->filename+strlen(r->filename)-4,"rcgi") == 0) ||
	    (strcmp(r->filename+strlen(r->filename)-5,"rhtml") == 0) ||
	    (strcmp(r->filename+strlen(r->filename)-5,"xhtml") == 0)) {
	    /* r->content_language=pstrdup(r->pool,"ru"); */
	    do_translate=1;
	}
#endif /* LOCAL_ROL_KLUDGE */
    }

/* now we know if we want to translate the text at all */

    if (!do_translate) {
#ifdef MCS_DEBUG
	printf("No translation for \"%s\",\"%s\"\n",
		r->content_language?r->content_language:"(null)",
		r->content_type?r->content_type:"(null)");
#endif
	return DECLINED;
    }

    do_translate=0;

/* First, check if the user said what charset she wants explicitly */

    strncpy(trname,r->server->server_hostname,MAX_TABNAME);
    trname[MAX_TABNAME]='\0';
    if ((p=strchr(trname,'.'))) *p='\0';

    ll=(table_entry *)charsets->elts;
    for (i=0;i<charsets->nelts;i++) {
	if (strcasecmp(ll[i].val,trname) == 0) {
#ifdef MCS_DEBUG
	    printf("server name matches \"%s\", set charset \"%s\"\n",
			ll[i].val,ll[i].key);
#endif
	    do_translate=1;
	    ccs=ll[i].key;
	    break;
	}
    }

/* If no 'translation' hostname, check if there was "Accept-Charset:" */

    if (!do_translate && acs) {
	char *acs2,*best=NULL,*q;
	float bestq=0.0,qual;
#ifdef MCS_DEBUG
	printf("accept-charset: \"%s\"\n",acs);
#endif
	/* This is a pain.  We need to parse charset string and choose best
	   match according to q= parameter... */
	acs2=pstrdup(r->pool,acs);
	for (p=strtok(acs2,",");p;p=strtok(NULL,",")) {
	    qual=1.0;
	    if ((q=strchr(p,';'))) {
		*q++='\0';
		while (isspace(*q)) q++;
		if (strncasecmp(q,"q=",2) == 0) {
		    if (sscanf(q+2,"%g",&qual) != 1) qual=1.0;
#ifdef MCS_DEBUG
		    printf("q= parameter \"%s\" gave %4.2f\n",q,qual);
#endif
		}
	    }
	    while (isspace(*p)) p++;
	    q=p+strlen(p)-1;
	    while ((q > p) && isspace(*q)) *q--='\0';
#ifdef MCS_DEBUG
	    printf("we have charset \"%s\" with quality %4.2f\n",p,qual);
#endif
	    if ((qual > bestq) && table_get(conf->charsets,p)) {
		bestq=qual;
		best=p;
	    }
	}
#ifdef MCS_DEBUG
	printf("best charset: \"%s\" with quality %4.2f\n",
		best?best:"(null)",bestq);
#endif
	if (best) {
	    p=table_get(conf->charsets,best);
#ifdef MCS_DEBUG
	    printf("we chose charset \"%s\" with table \"%s\"\n",best,p);
#endif
	    strncpy(trname,p,MAX_TABNAME);
	    trname[3]='\0';
	    do_translate=1;
	    ccs=best;
	}
    }

/* As the last resort, try determine desired charset by the user agent */

    if (!do_translate) {
	char *ua=table_get(r->headers_in,"User-Agent");
	if (ua) {
	    ll=(table_entry*)agents->elts;
#ifdef MCS_DEBUG
	    printf("check user-agent \"%s\"\n",ua);
#endif
		/* We cannot make simple table_get here because we want to
		   check strstr() match rather than strict match... */
	    for (i=0;i<agents->nelts;i++) {
		if (strstr(ua,ll[i].key)) {
#ifdef MCS_DEBUG
		    printf("user agent matches allowed table \"%s\"\n",
				ll[i].key);
#endif
		    do_translate=1;
		    strncpy(trname,ll[i].val,3);
		    trname[3]='\0';
		    break;
		}
	    }
	}
    }

    if (do_translate) {
	if (!ccs) {
	    ll=(table_entry*)charsets->elts;
	    for (i=0;i<charsets->nelts;i++) {
		if (strcasecmp(trname,ll[i].val) == 0) {
		    ccs=ll[i].key;
		    break;
		}
	    }
	}
#ifdef MCS_DEBUG
	printf("adding charset \"%s\"\n",ccs?ccs:"(null)");
#endif
	if (ccs) r->content_charset=pstrdup(r->pool,ccs);
	table_set (e, "LANG",trname);
	table_set (e, "CHARSET_TRANSLATION",trname);
#ifdef MCS_DEBUG
	printf("use \"%s\"\n",trname);
#endif
	btranslate=translate_buffer;
	i=settrtab_byname(trname);
#ifdef MCS_DEBUG
	printf("settrtab(%s) rc=%d\n",trname,i);
#endif
    }

    return DECLINED;
}

module mcs_module = {
   STANDARD_MODULE_STUFF,
   initialize_mcs,		/* initializer */
   create_mcs_config,		/* dir config creater */
   merge_mcs_config,		/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server configs */
   mcs_cmds,			/* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   NULL,			/* check_user_id */
   NULL,			/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   setup_translation,		/* fixups */
   NULL				/* logger */
};
